# coding: utf-8
#

import re
import time

from six.moves import shlex_quote
import uiautomator2
from logzero import logger
from lxml import etree


def safe_xmlstr(s):
    return s.replace("$", "-")


def init():
    uiautomator2.plugin_register("xpath", XPath)


class TimeoutException(Exception):
    pass


class XPath(object):
    def __init__(self, d):
        """
        Args:
            d (uiautomator2 instance)
        """
        self._d = d
        self._watchers = []  # item: {"xpath": .., "callback": func}
        self._timeout = 10.0

    def implicitly_wait(self, timeout):
        """ set default timeout when click """
        self._timeout = timeout

    def match(self, xpath, source=None):
        return len(self(xpath, source).all()) > 0

    def when(self, xpath):
        obj = self

        def _click(selector):
            selector.click_nowait()

        class _Watcher():
            def click(self):
                obj._watchers.append({
                    "xpath": xpath,
                    "callback": _click,
                })

            def call(self, func):
                """
                Args:
                    func: accept only one argument "selector"
                """
                obj._watchers.append({
                    "xpath": xpath,
                    "callback": func,
                })

        return _Watcher()

    def run_watchers(self, source=None):
        source = source or self._d.dump_hierarchy()
        for h in self._watchers:
            selector = self(h['xpath'], source)
            if selector.exists:
                logger.info("XPath(hook) %s", h['xpath'])
                h['callback'](selector)
                return True
        return False

    def sleep_watch(self, seconds):
        """ run watchers when sleep """
        deadline = time.time() + seconds
        while time.time() < deadline:
            self.run_watchers()
            left_time = max(0, deadline - time.time())
            time.sleep(min(0.5, left_time))

    def click(self, xpath, source=None, watch=True, timeout=None):
        timeout = timeout or self._timeout
        logger.info("XPath(timeout %.1f) %s", timeout, xpath)

        deadline = time.time() + timeout
        while time.time() < deadline:
            source = self._d.dump_hierarchy()
            if watch and self.run_watchers(source):
                time.sleep(.5)  # post delay
                continue

            selector = self(xpath, source)
            if selector.exists:
                selector.click_nowait()
                time.sleep(.5)  # post sleep
                return
            time.sleep(.5)
            # source = self._d.dump_hierarchy()
        raise TimeoutException("timeout %.1f" % timeout)

    def __call__(self, xpath, source=None):
        if xpath.startswith('//'):
            pass
        elif xpath.startswith('@'):
            xpath = '//*[@resource-id={}]'.format(shlex_quote(xpath[1:]))
        elif xpath.startswith('^'):
            xpath = '//*[re:match(text(), {})]'.format(shlex_quote(xpath))
        elif xpath.startswith('%') and xpath.endswith("%"):
            xpath = '//*[contains(text(), {}]'.format(shlex_quote(xpath))
        elif xpath.startswith('%'):
            xpath = '//*[starts-with(text(), {}]'.format(shlex_quote(xpath))
        elif xpath.endswith('%'):
            xpath = '//*[ends-with(text(), {}]'.format(shlex_quote(xpath))
        else:
            xpath = '//*[@text={0} or @content-desc={0}]'.format(
                shlex_quote(xpath))
        return XPathSelector(self, xpath, source)


class XPathSelector(object):
    def __init__(self, parent, xpath, source=None):
        self._parent = parent
        self._d = parent._d
        self._xpath = xpath
        self._source = source
        self._watchers = []

    def all(self):
        """
        Returns:
            list of XMLElement
        """
        xml_content = self._source or self._d.dump_hierarchy()
        root = etree.fromstring(xml_content.encode('utf-8'))
        for node in root.xpath("//node"):
            node.tag = safe_xmlstr(node.attrib.pop("class"))
        match_nodes = root.xpath(
            self._xpath,
            namespaces={"re": "http://exslt.org/regular-expressions"})
        return [XMLElement(node) for node in match_nodes]

    @property
    def exists(self):
        return len(self.all()) > 0

    def click_nowait(self):
        x, y = self.all()[0].center()
        logger.info("click %d, %d", x, y)
        self._d.click(x, y)

    def click(self, timeout=None):
        self._parent.click(self._xpath, timeout=timeout)


class XMLElement(object):
    def __init__(self, elem):
        self.elem = elem

    def center(self):
        bounds = self.elem.attrib.get("bounds")
        lx, ly, rx, ry = map(int, re.findall(r"\d+", bounds))
        return (lx + rx) // 2, (ly + ry) // 2

    @property
    def text(self):
        return self.elem.attrib.get("text")

    @property
    def attrib(self):
        return self.elem.attrib


if __name__ == "__main__":
    init()
    import uiautomator2.ext.htmlreport as htmlreport

    d = uiautomator2.connect()
    hrp = htmlreport.HTMLReport(d)

    # take screenshot before each click
    hrp.patch_click()
    d.app_start("com.netease.cloudmusic", stop=True)

    # watchers
    d.ext_xpath.when("跳过").click()
    # d.ext_xpath.when("知道了").click()

    # steps
    d.ext_xpath("//*[@text='私人FM']/../android.widget.ImageView").click()
    d.ext_xpath("下一首").click()
    d.ext_xpath.sleep_watch(2)
    d.ext_xpath("转到上一层级").click()


# coding: utf-8
#
# Test apk Download from
# https://github.com/appium/java-client/raw/master/src/test/java/io/appium/java_client/ApiDemos-debug.apk

import uiautomator2 as u2
import unittest
import time


class SimpleTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.d = u2.connect()
        cls.d.set_orientation('natural')
        cls.d.implicitly_wait(10)

    def setUp(self):
        self.sess = self.d.session("io.appium.android.apis")
        self.sess.watchers.remove()

    def tearDown(self):
        self.sess.watchers.remove()

    def test_toast_get_message(self):
        d = self.sess
        d.toast.reset()
        assert d.toast.get_message(0) is None
        assert d.toast.get_message(0, default="d") == "d"
        d(text="App").click()
        d(text="Notification").click()
        d(text="NotifyWithText").click()
        try:
            d(text="Show Short Notification").click()
        except u2.UiObjectNotFoundError:
            d(text="SHOW SHORT NOTIFICATION").click()
        self.assertEqual(d.toast.get_message(2, 5, ""), "Short notification")
        time.sleep(.5)
        self.assertIsNone(d.toast.get_message(0, 0.4))
        # d.toast.reset()
        # d.toast.show("Hello world")
        # self.assertEqual(d.toast.get_message(5, 5), "Hello world")

    def test_scroll(self):
        d = self.sess
        d(text="App").click()
        d(scrollable=True).scroll.to(text="Voice Recognition")

    def test_watchers(self):
        """
        App -> Notification -> Status Bar
        """
        d = self.sess
        d.watchers.remove()
        d.watchers.watched = False

        d.watcher("N").when(text='Notification').click(text='Notification')
        d(text="App").click()
        d.watchers.run()
        self.assertTrue(d(text="Status Bar").wait(timeout=3))
        d.press("back")
        d.press("back")
        # Should auto click Notification when show up
        self.assertFalse(d.watchers.watched)
        d.watchers.watched = True
        self.assertTrue(d.watchers.watched)
        d(text="App").click()
        self.assertTrue(d(text="Status Bar").exists(timeout=3))
        d.watchers.watched = False

    def test_count(self):
        d = self.sess
        count = d(resourceId="android:id/list").child(
            className="android.widget.TextView").count
        self.assertEqual(count, 11)
        self.assertEqual(
            d(resourceId="android:id/list").info['childCount'], 11)
        count = d(resourceId="android:id/list").child(
            className="android.widget.TextView", instance=0).count
        self.assertEqual(count, 1)

    def test_get_text(self):
        d = self.sess
        text = d(resourceId="android:id/list").child(
            className="android.widget.TextView", instance=2).get_text()
        self.assertEqual(text, "App")

    def test_xpath(self):
        d = self.sess
        self.assertEqual(len(d.xpath("//*[@text='Media']").all()), 1)
        self.assertEqual(len(d.xpath("//*[@text='MediaNotExists']").all()), 0)
        d.xpath("//*[@text='Media']").click()
        self.assertTrue(d.xpath('//*[contains(@text, "Audio")]').wait(5))

    def test_implicitly_wait(self):
        d = self.sess
        d.implicitly_wait(2)
        self.assertEqual(d.implicitly_wait(), 2)
        start = time.time()
        with self.assertRaises(u2.UiObjectNotFoundError):
            d(text="Sensors").get_text()
        time_used = time.time() - start
        self.assertGreaterEqual(time_used, 2)
        # maybe longer then 2, waitForExists -> getText
        # getText may take 1~2s
        self.assertLess(time_used, 2 + 3)

    def test_select_iter(self):
        d = self.sess
        d(text='OS').click()
        texts = d(resourceId='android:id/list').child(
            className='android.widget.TextView')
        self.assertEqual(texts.count, 4)
        words = []
        for item in texts:
            words.append(item.get_text())
        self.assertEqual(
            words,
            ['Morse Code', 'Rotation Vector', 'Sensors', 'SMS Messaging'])

    def test_plugin(self):
        def _my_plugin(d, k):
            def _inner():
                return k

            return _inner

        u2.plugin_clear()
        u2.plugin_register('my', _my_plugin, 'pp')
        self.assertEqual(self.d.ext_my(), 'pp')


if __name__ == '__main__':
    unittest.main()
